package namesys

import (
	"context"
	"testing"
	"time"

	ds "gitlab.dms3.io/dms3/go-datastore"
	dssync "gitlab.dms3.io/dms3/go-datastore/sync"
	mockrouting "gitlab.dms3.io/dms3/go-dms3-routing/mock"
	offline "gitlab.dms3.io/dms3/go-dms3-routing/offline"
	dms3ns "gitlab.dms3.io/dms3/go-dms3ns"
	dms3ns_pb "gitlab.dms3.io/dms3/go-dms3ns/pb"
	path "gitlab.dms3.io/dms3/go-path"
	opts "gitlab.dms3.io/dms3/interface-go-dms3-core/options/namesys"
	ci "gitlab.dms3.io/p2p/go-p2p-core/crypto"
	peer "gitlab.dms3.io/p2p/go-p2p-core/peer"
	pstore "gitlab.dms3.io/p2p/go-p2p-core/peerstore"
	routing "gitlab.dms3.io/p2p/go-p2p-core/routing"
	"gitlab.dms3.io/p2p/go-p2p-core/test"
	pstoremem "gitlab.dms3.io/p2p/go-p2p-peerstore/pstoremem"
	record "gitlab.dms3.io/p2p/go-p2p-record"
	testutil "gitlab.dms3.io/p2p/go-p2p-testing/net"
)

func TestResolverValidation(t *testing.T) {
	t.Run("RSA",
		func(t *testing.T) {
			testResolverValidation(t, ci.RSA)
		})
	t.Run("Ed25519",
		func(t *testing.T) {
			testResolverValidation(t, ci.Ed25519)
		})
	t.Run("ECDSA",
		func(t *testing.T) {
			testResolverValidation(t, ci.ECDSA)
		})
	t.Run("Secp256k1",
		func(t *testing.T) {
			testResolverValidation(t, ci.Secp256k1)
		})
}

func testResolverValidation(t *testing.T, keyType int) {
	ctx := context.Background()
	rid := testutil.RandIdentityOrFatal(t)
	dstore := dssync.MutexWrap(ds.NewMapDatastore())
	peerstore := pstoremem.NewPeerstore()

	vstore := newMockValueStore(rid, dstore, peerstore)
	resolver := NewDms3NsResolver(vstore)

	nvVstore := offline.NewOfflineRouter(dstore, mockrouting.MockValidator{})

	// Create entry with expiry in one hour
	priv, id, _, dms3nsDHTPath := genKeys(t, keyType)
	ts := time.Now()
	p := []byte("/dms3/QmfM2r8seH2GiRaC4esTjeraXEachRt8ZsSeGaWTPLyMoG")
	entry, err := createDMS3NSRecordWithEmbeddedPublicKey(priv, p, 1, ts.Add(time.Hour))
	if err != nil {
		t.Fatal(err)
	}

	// Publish entry
	err = PublishEntry(ctx, vstore, dms3nsDHTPath, entry)
	if err != nil {
		t.Fatal(err)
	}

	// Resolve entry
	resp, err := resolve(ctx, resolver, id.Pretty(), opts.DefaultResolveOpts())
	if err != nil {
		t.Fatal(err)
	}
	if resp != path.Path(p) {
		t.Fatalf("Mismatch between published path %s and resolved path %s", p, resp)
	}
	// Create expired entry
	expiredEntry, err := createDMS3NSRecordWithEmbeddedPublicKey(priv, p, 1, ts.Add(-1*time.Hour))
	if err != nil {
		t.Fatal(err)
	}

	// Publish entry
	err = PublishEntry(ctx, nvVstore, dms3nsDHTPath, expiredEntry)
	if err != nil {
		t.Fatal(err)
	}

	// Record should fail validation because entry is expired
	_, err = resolve(ctx, resolver, id.Pretty(), opts.DefaultResolveOpts())
	if err == nil {
		t.Fatal("ValidateDms3NsRecord should have returned error")
	}

	// Create DMS3NS record path with a different private key
	priv2, id2, _, dms3nsDHTPath2 := genKeys(t, keyType)

	// Publish entry
	err = PublishEntry(ctx, nvVstore, dms3nsDHTPath2, entry)
	if err != nil {
		t.Fatal(err)
	}

	// Record should fail validation because public key defined by
	// dms3ns path doesn't match record signature
	_, err = resolve(ctx, resolver, id2.Pretty(), opts.DefaultResolveOpts())
	if err == nil {
		t.Fatal("ValidateDms3NsRecord should have failed signature verification")
	}

	// Try embedding the incorrect private key inside the entry
	if err := dms3ns.EmbedPublicKey(priv2.GetPublic(), entry); err != nil {
		t.Fatal(err)
	}

	// Publish entry
	err = PublishEntry(ctx, nvVstore, dms3nsDHTPath2, entry)
	if err != nil {
		t.Fatal(err)
	}

	// Record should fail validation because public key defined by
	// dms3ns path doesn't match record signature
	_, err = resolve(ctx, resolver, id2.Pretty(), opts.DefaultResolveOpts())
	if err == nil {
		t.Fatal("ValidateDms3NsRecord should have failed signature verification")
	}
}

func genKeys(t *testing.T, keyType int) (ci.PrivKey, peer.ID, string, string) {
	bits := 0
	if keyType == ci.RSA {
		bits = 2048
	}

	sk, pk, err := test.RandTestKeyPair(keyType, bits)
	if err != nil {
		t.Fatal(err)
	}
	id, err := peer.IDFromPublicKey(pk)
	if err != nil {
		t.Fatal(err)
	}
	return sk, id, PkKeyForID(id), dms3ns.RecordKey(id)
}

func createDMS3NSRecordWithEmbeddedPublicKey(sk ci.PrivKey, val []byte, seq uint64, eol time.Time) (*dms3ns_pb.Dms3NsEntry, error) {
	entry, err := dms3ns.Create(sk, val, seq, eol)
	if err != nil {
		return nil, err
	}
	if err := dms3ns.EmbedPublicKey(sk.GetPublic(), entry); err != nil {
		return nil, err
	}

	return entry, nil
}

type mockValueStore struct {
	r     routing.ValueStore
	kbook pstore.KeyBook
}

func newMockValueStore(id testutil.Identity, dstore ds.Datastore, kbook pstore.KeyBook) *mockValueStore {
	return &mockValueStore{
		r: offline.NewOfflineRouter(dstore, record.NamespacedValidator{
			"dms3ns": dms3ns.Validator{KeyBook: kbook},
			"pk":     record.PublicKeyValidator{},
		}),
		kbook: kbook,
	}
}

func (m *mockValueStore) GetValue(ctx context.Context, k string, opts ...routing.Option) ([]byte, error) {
	return m.r.GetValue(ctx, k, opts...)
}

func (m *mockValueStore) SearchValue(ctx context.Context, k string, opts ...routing.Option) (<-chan []byte, error) {
	return m.r.SearchValue(ctx, k, opts...)
}

func (m *mockValueStore) GetPublicKey(ctx context.Context, p peer.ID) (ci.PubKey, error) {
	pk := m.kbook.PubKey(p)
	if pk != nil {
		return pk, nil
	}

	pkkey := routing.KeyForPublicKey(p)
	val, err := m.GetValue(ctx, pkkey)
	if err != nil {
		return nil, err
	}

	pk, err = ci.UnmarshalPublicKey(val)
	if err != nil {
		return nil, err
	}

	return pk, m.kbook.AddPubKey(p, pk)
}

func (m *mockValueStore) PutValue(ctx context.Context, k string, d []byte, opts ...routing.Option) error {
	return m.r.PutValue(ctx, k, d, opts...)
}
