// Package republisher provides a utility to automatically re-publish DMS3NS
// records related to the keys in a Keystore.
package republisher

import (
	"context"
	"errors"
	"time"

	keystore "gitlab.dms3.io/dms3/go-dms3-keystore"
	namesys "gitlab.dms3.io/dms3/go-namesys"
	path "gitlab.dms3.io/dms3/go-path"

	proto "github.com/gogo/protobuf/proto"
	goprocess "github.com/jbenet/goprocess"
	gpctx "github.com/jbenet/goprocess/context"
	ds "gitlab.dms3.io/dms3/go-datastore"
	dms3ns "gitlab.dms3.io/dms3/go-dms3ns"
	pb "gitlab.dms3.io/dms3/go-dms3ns/pb"
	logging "gitlab.dms3.io/dms3/go-log"
	ic "gitlab.dms3.io/p2p/go-p2p-core/crypto"
	peer "gitlab.dms3.io/p2p/go-p2p-core/peer"
)

var errNoEntry = errors.New("no previous entry")

var log = logging.Logger("dms3ns-repub")

// DefaultRebroadcastInterval is the default interval at which we rebroadcast DMS3NS records
var DefaultRebroadcastInterval = time.Hour * 4

// InitialRebroadcastDelay is the delay before first broadcasting DMS3NS records on start
var InitialRebroadcastDelay = time.Minute * 1

// FailureRetryInterval is the interval at which we retry DMS3NS records broadcasts (when they fail)
var FailureRetryInterval = time.Minute * 5

// DefaultRecordLifetime is the default lifetime for DMS3NS records
const DefaultRecordLifetime = time.Hour * 24

// Republisher facilitates the regular publishing of all the DMS3NS records
// associated to keys in a Keystore.
type Republisher struct {
	ns   namesys.Publisher
	ds   ds.Datastore
	self ic.PrivKey
	ks   keystore.Keystore

	Interval time.Duration

	// how long records that are republished should be valid for
	RecordLifetime time.Duration
}

// NewRepublisher creates a new Republisher
func NewRepublisher(ns namesys.Publisher, ds ds.Datastore, self ic.PrivKey, ks keystore.Keystore) *Republisher {
	return &Republisher{
		ns:             ns,
		ds:             ds,
		self:           self,
		ks:             ks,
		Interval:       DefaultRebroadcastInterval,
		RecordLifetime: DefaultRecordLifetime,
	}
}

// Run starts the republisher facility. It can be stopped by stopping the
// provided proc.
func (rp *Republisher) Run(proc goprocess.Process) {
	timer := time.NewTimer(InitialRebroadcastDelay)
	defer timer.Stop()
	if rp.Interval < InitialRebroadcastDelay {
		timer.Reset(rp.Interval)
	}

	for {
		select {
		case <-timer.C:
			timer.Reset(rp.Interval)
			err := rp.republishEntries(proc)
			if err != nil {
				log.Info("republisher failed to republish: ", err)
				if FailureRetryInterval < rp.Interval {
					timer.Reset(FailureRetryInterval)
				}
			}
		case <-proc.Closing():
			return
		}
	}
}

func (rp *Republisher) republishEntries(p goprocess.Process) error {
	ctx, cancel := context.WithCancel(gpctx.OnClosingContext(p))
	defer cancel()

	// TODO: Use rp.dms3ns.ListPublished(). We can't currently *do* that
	// because:
	// 1. There's no way to get keys from the keystore by ID.
	// 2. We don't actually have access to the DMS3NS publisher.
	err := rp.republishEntry(ctx, rp.self)
	if err != nil {
		return err
	}

	if rp.ks != nil {
		keyNames, err := rp.ks.List()
		if err != nil {
			return err
		}
		for _, name := range keyNames {
			priv, err := rp.ks.Get(name)
			if err != nil {
				return err
			}
			err = rp.republishEntry(ctx, priv)
			if err != nil {
				return err
			}

		}
	}

	return nil
}

func (rp *Republisher) republishEntry(ctx context.Context, priv ic.PrivKey) error {
	id, err := peer.IDFromPrivateKey(priv)
	if err != nil {
		return err
	}

	log.Debugf("republishing dms3ns entry for %s", id)

	// Look for it locally only
	e, err := rp.getLastDMS3NSEntry(id)
	if err != nil {
		if err == errNoEntry {
			return nil
		}
		return err
	}

	p := path.Path(e.GetValue())
	prevEol, err := dms3ns.GetEOL(e)
	if err != nil {
		return err
	}

	// update record with same sequence number
	eol := time.Now().Add(rp.RecordLifetime)
	if prevEol.After(eol) {
		eol = prevEol
	}
	return rp.ns.PublishWithEOL(ctx, priv, p, eol)
}

func (rp *Republisher) getLastDMS3NSEntry(id peer.ID) (*pb.Dms3NsEntry, error) {
	// Look for it locally only
	val, err := rp.ds.Get(namesys.Dms3NsDsKey(id))
	switch err {
	case nil:
	case ds.ErrNotFound:
		return nil, errNoEntry
	default:
		return nil, err
	}

	e := new(pb.Dms3NsEntry)
	if err := proto.Unmarshal(val, e); err != nil {
		return nil, err
	}
	return e, nil
}
