package namesys

import (
	"context"
	"strings"
	"time"

	proto "github.com/gogo/protobuf/proto"
	mh "github.com/multiformats/go-multihash"
	cid "gitlab.dms3.io/dms3/go-cid"
	dms3ns "gitlab.dms3.io/dms3/go-dms3ns"
	pb "gitlab.dms3.io/dms3/go-dms3ns/pb"
	logging "gitlab.dms3.io/dms3/go-log"
	path "gitlab.dms3.io/dms3/go-path"
	opts "gitlab.dms3.io/dms3/interface-go-dms3-core/options/namesys"
	peer "gitlab.dms3.io/p2p/go-p2p-core/peer"
	routing "gitlab.dms3.io/p2p/go-p2p-core/routing"
	dht "gitlab.dms3.io/p2p/go-p2p-kad-dht"
)

var log = logging.Logger("namesys")

// Dms3NsResolver implements NSResolver for the main DMS3 SFS-like naming
type Dms3NsResolver struct {
	routing routing.ValueStore
}

// NewDms3NsResolver constructs a name resolver using the DMS3 Routing system
// to implement SFS-like naming on top.
func NewDms3NsResolver(route routing.ValueStore) *Dms3NsResolver {
	if route == nil {
		panic("attempt to create resolver with nil routing system")
	}
	return &Dms3NsResolver{
		routing: route,
	}
}

// Resolve implements Resolver.
func (r *Dms3NsResolver) Resolve(ctx context.Context, name string, options ...opts.ResolveOpt) (path.Path, error) {
	return resolve(ctx, r, name, opts.ProcessOpts(options))
}

// ResolveAsync implements Resolver.
func (r *Dms3NsResolver) ResolveAsync(ctx context.Context, name string, options ...opts.ResolveOpt) <-chan Result {
	return resolveAsync(ctx, r, name, opts.ProcessOpts(options))
}

// resolveOnce implements resolver. Uses the DMS3 routing system to
// resolve SFS-like names.
func (r *Dms3NsResolver) resolveOnceAsync(ctx context.Context, name string, options opts.ResolveOpts) <-chan onceResult {
	out := make(chan onceResult, 1)
	log.Debugf("RoutingResolver resolving %s", name)
	cancel := func() {}

	if options.DhtTimeout != 0 {
		// Resolution must complete within the timeout
		ctx, cancel = context.WithTimeout(ctx, options.DhtTimeout)
	}

	name = strings.TrimPrefix(name, "/dms3ns/")

	pid, err := peer.Decode(name)
	if err != nil {
		log.Debugf("RoutingResolver: could not convert public key hash %s to peer ID: %s\n", name, err)
		out <- onceResult{err: err}
		close(out)
		cancel()
		return out
	}

	// Use the routing system to get the name.
	// Note that the DHT will call the dms3ns validator when retrieving
	// the value, which in turn verifies the dms3ns record signature
	dms3nsKey := dms3ns.RecordKey(pid)

	vals, err := r.routing.SearchValue(ctx, dms3nsKey, dht.Quorum(int(options.DhtRecordCount)))
	if err != nil {
		log.Debugf("RoutingResolver: dht get for name %s failed: %s", name, err)
		out <- onceResult{err: err}
		close(out)
		cancel()
		return out
	}

	go func() {
		defer cancel()
		defer close(out)
		for {
			select {
			case val, ok := <-vals:
				if !ok {
					return
				}

				entry := new(pb.Dms3NsEntry)
				err = proto.Unmarshal(val, entry)
				if err != nil {
					log.Debugf("RoutingResolver: could not unmarshal value for name %s: %s", name, err)
					emitOnceResult(ctx, out, onceResult{err: err})
					return
				}

				var p path.Path
				// check for old style record:
				if valh, err := mh.Cast(entry.GetValue()); err == nil {
					// Its an old style multihash record
					log.Debugf("encountered CIDv0 dms3ns entry: %s", valh)
					p = path.FromCid(cid.NewCidV0(valh))
				} else {
					// Not a multihash, probably a new style record
					p, err = path.ParsePath(string(entry.GetValue()))
					if err != nil {
						emitOnceResult(ctx, out, onceResult{err: err})
						return
					}
				}

				ttl := DefaultResolverCacheTTL
				if entry.Ttl != nil {
					ttl = time.Duration(*entry.Ttl)
				}
				switch eol, err := dms3ns.GetEOL(entry); err {
				case dms3ns.ErrUnrecognizedValidity:
					// No EOL.
				case nil:
					ttEol := time.Until(eol)
					if ttEol < 0 {
						// It *was* valid when we first resolved it.
						ttl = 0
					} else if ttEol < ttl {
						ttl = ttEol
					}
				default:
					log.Errorf("encountered error when parsing EOL: %s", err)
					emitOnceResult(ctx, out, onceResult{err: err})
					return
				}

				emitOnceResult(ctx, out, onceResult{value: p, ttl: ttl})
			case <-ctx.Done():
				return
			}
		}
	}()

	return out
}
